import pytest

from jptest2 import PythonNotebook


@pytest.mark.asyncio
async def test_replace_function():
    def local_fun(_, b):
        return b

    async with PythonNotebook('functions.ipynb') as nb:
        # execute cell with function definition
        await nb.execute_cells('definition')

        # replace function
        async with nb.replace_fun('nb_fun', local_fun):
            # compare results
            cells = await nb.execute_cells('execution')
            result, *_ = cells[0].output()

            assert any([val == '10' for _, val in result])

            await nb.execute_cells('store')
            result = await nb.ref('result').receive()
            assert result == 110

        # get results again after backup was restored
        cells = await nb.execute_cells('execution')
        result, *_ = cells[0].output()

        assert any([val == '5' for _, val in result])

        await nb.execute_cells('store')
        result = await nb.ref('result').receive()
        assert result == 105


@pytest.mark.asyncio
async def test_replace_module_function():
    def local_fun(*args, **kwargs):
        return 1, 2, 3

    async with PythonNotebook('functions.ipynb') as nb:
        # import module
        await nb.execute_code('''
            import math
        ''')

        # replace function in module
        async with nb.replace_fun('math.ceil', local_fun):
            # get function and receive result
            result = await nb.ref('math.ceil')().receive()
            assert result == (1, 2, 3)


@pytest.mark.asyncio
async def test_replace_class_function():
    def local_fun(*args, **kwargs):
        return 1, 2, 3

    async with PythonNotebook('functions.ipynb') as nb:
        # inject class
        await nb.execute_code('''
            class Test:
                def fun(self):
                    return 5
        ''')

        # replace function
        async with nb.replace_fun('Test.fun', local_fun):
            # create object and receive result
            await nb.execute_code('a = Test()')
            result = await nb.ref('a').fun().receive()

            assert result == (1, 2, 3)


@pytest.mark.asyncio
async def test_replace_object_function():
    def local_fun(*args, **kwargs):
        return 1, 2, 3

    async with PythonNotebook('functions.ipynb') as nb:
        # inject class
        await nb.execute_code('''
            class Test:
                def fun(self):
                    return 5
                    
            a = Test()
            b = Test()
        ''')

        # replace function
        async with nb.replace_fun('a.fun', local_fun):
            result_a = await nb.ref('a').fun().receive()
            assert result_a == (1, 2, 3)

            result_b = await nb.ref('b').fun().receive()
            assert result_b == 5



from typing import Callable

from . import Notebook
from . import NotebookReference


class NotebookFunctionReplacement:
    def __init__(self, nb: Notebook, fun_name: str, replacement: Callable):
        self._nb: Notebook = nb
        self._fun_name = fun_name
        self._replacement: Callable = replacement

    async def __aenter__(self) -> "NotebookFunctionReplacement":
        """
        replace function with replacement

        :return: self
        """
        self._backup: NotebookReference = await self._nb.get(self._fun_name)
        injected = await (await self._nb.inject_fun(self._replacement)).copy()

        await self._nb.execute_code(f'''
            {self._fun_name} = {injected.name}
        ''')

        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        """
        replace replacement with backup

        :param exc_type:
        :param exc_val:
        :param exc_tb:
        :return:
        """
        await self._nb.execute_code(f'''
            {self._fun_name} = {self._backup.name}
        ''')



import asyncio

import pytest

from jptest2 import PythonNotebook


@pytest.mark.asyncio
async def test_receive():
    async with PythonNotebook('references.ipynb') as nb:
        await nb.execute_cells('primitives', 'objects')

        nb_int = await nb.ref('nb_int').receive()
        assert nb_int == 1024

        nb_float = await nb.ref('nb_float').receive()
        assert nb_float == 5.25

        nb_str = await nb.ref('nb_str').receive()
        assert nb_str == 'text'

        nb_bool = await nb.ref('nb_bool').receive()
        assert nb_bool is True

        nb_obj = await nb.ref('nb_obj').receive()
        assert nb_obj.key == 'val'

        nb_list = await nb.ref('nb_list').receive()
        assert nb_list == [1, 'a', None, 'b']

        nb_tuple = await nb.ref('nb_tuple').receive()
        assert nb_tuple == (2, 'b')

        nb_set = await nb.ref('nb_set').receive()
        assert nb_set == {5, 3, 1}

        nb_dict = await nb.ref('nb_dict').receive()
        assert nb_dict == {
            'a': nb_obj,
            'b': 2
        }


@pytest.mark.asyncio
async def test_reference_call():
    # receive
    async with PythonNotebook('references.ipynb') as nb:
        await nb.execute_cells('create', 'objects')

        nb_fun = nb.ref('nb_swap')
        nb_a, nb_b, nb_c = nb.refs('a', 'b', 'c')

        lo_a = await nb_a.receive()
        lo_b = await nb_b.receive()
        lo_c = await nb_c.receive()

        # args: call with PythonNotebook parameters
        result = await nb_fun(nb_a, nb_c).receive()
        assert result == (lo_c, lo_a)

        # args: call with local parameters
        result = await nb_fun(lo_a, lo_c).receive()
        assert result == (lo_c, lo_a)

        # args: call with mixed parameters
        result = await nb_fun(nb_a, lo_c).receive()
        assert result == (lo_c, lo_a)

        # kwargs: call with PythonNotebook parameter
        result = await nb_fun(nb_a, nb_c, replace_second=nb_b).receive()
        assert result == (lo_b, lo_a)

        # kwargs: call with local parameter
        result = await nb_fun(nb_a, nb_c, replace_second=lo_b).receive()
        assert result == (lo_b, lo_a)

    # execute
    async with PythonNotebook('references.ipynb') as nb:
        await nb.execute_cells('create', 'objects')

        nb_fun = nb.ref('nb_swap')
        nb_a, nb_b, nb_c = nb.refs('a', 'b', 'c')
        lo_a, lo_b, lo_c = 1, 2, 3

        # args: call with PythonNotebook parameters
        await nb_fun(nb_a, nb_c).execute()

        # args: call with local parameters
        await nb_fun(lo_a, lo_c).execute()

        # args: call with mixed parameters
        await nb_fun(nb_a, lo_c).execute()

        # kwargs: call with PythonNotebook parameter
        await nb_fun(nb_a, nb_c, replace_second=nb_b).execute()

        # kwargs: call with local parameter
        await nb_fun(nb_a, nb_c, replace_second=lo_b).execute()


@pytest.mark.asyncio
async def test_item_and_attr():
    async with PythonNotebook('references.ipynb') as nb:
        await nb.execute_cells('objects')

        # item
        nb_dict_b = await nb.ref('nb_dict')['b'].receive()
        assert nb_dict_b == 2

        # attr
        nb_obj_key = await nb.ref('nb_obj').key.receive()
        assert nb_obj_key == 'val'

        # mixed
        nb_mixed = await nb.ref('nb_dict')['a'].key.receive()
        assert nb_mixed == 'val'


@pytest.mark.asyncio
async def test_len():
    async with PythonNotebook('references.ipynb') as nb:
        await nb.execute_cells('objects')

        nb_list_len = await nb.ref('nb_list').len()
        assert nb_list_len == 4


@pytest.mark.asyncio
async def test_copy():
    async with PythonNotebook('references.ipynb') as nb:
        await nb.execute_cells('objects')

        # simple name
        nb_list = await nb.ref('nb_list').copy()
        await nb.execute_code('nb_list = 3.142')

        r1 = await nb_list.receive()
        assert r1 == [1, 'a', None, 'b']

        r2 = await nb.ref('nb_list').receive()
        assert r2 == 3.142

        # property
        nb_obj = await nb.ref('nb_obj').key.copy()
        await nb.execute_code('nb_obj.key = "changed"')

        r3 = await nb_obj.receive()
        assert r3 == 'val'

        r4 = await nb.ref('nb_obj').key.receive()
        assert r4 == 'changed'


@pytest.mark.asyncio
async def test_transfer_between_notebooks():
    async with \
            PythonNotebook('references.ipynb') as nb1, \
            PythonNotebook('references.ipynb') as nb2:
        await asyncio.gather(
            nb1.execute_cells('create', 'primitives'),
            nb2.execute_cells('objects')
        )

        nb1_fun = nb1.ref('nb_swap')

        # pick a var from `nb2` that is _not_ available in `nb1`
        nb1_a, nb2_obj = nb1.ref('a'), nb2.ref('nb_obj')

        result = await nb1_fun(nb1_a, nb2_obj).receive()
        assert result == (await nb2_obj.receive(), await nb1_a.receive())

        # pick a var from `nb2` that is available in `nb1`
        await nb2.execute_code('nb_int = 3072')

        nb1_a, nb2_int = nb1.ref('a'), nb2.ref('nb_int')

        result = await nb1_fun(nb1_a, nb2_int).receive()
        assert result == (await nb2_int.receive(), await nb1_a.receive())



import asyncio
from os import PathLike
from typing import Union, List, Optional

from .JPTest import JPTest, EXECUTE_TYPE
from .notebook.kernels import PythonNotebook


class JPTestComparison(JPTest):
    """
    decorator to use with test functions
    """

    def __init__(self, name: str = None, max_score: Union[float, int] = None, timeout: int = 120,
                 prepare: EXECUTE_TYPE = None, execute_left: EXECUTE_TYPE = None, execute_right: EXECUTE_TYPE = None,
                 hold_left: Union[str, List[str]] = None, hold_right: Union[str, List[str]] = None):
        super().__init__(name, max_score, timeout, prepare)

        self._execute_left: EXECUTE_TYPE = execute_left if execute_left is not None else []
        self._execute_right: EXECUTE_TYPE = execute_right if execute_right is not None else []
        self._hold_left: Optional[Union[str, List[str]]] = hold_left
        self._hold_right: Optional[Union[str, List[str]]] = hold_right

    async def __prepare(self, nb: PythonNotebook, prepare: EXECUTE_TYPE, execute: EXECUTE_TYPE, hold: List[str]):
        await self._execute_recursively(nb, prepare)
        await self._execute_recursively(nb, execute)

        refs = nb.refs(*hold)
        vals = await asyncio.gather(*[ref.receive() for ref in refs])

        return vals

    async def execute(self, notebook: Union[str, PathLike]):
        async with \
                self._start(notebook) as left, \
                self._start(notebook) as right:
            # hold
            hold_left = self._hold_left if isinstance(self._hold_left, list) else [self._hold_left]
            hold_right = self._hold_right if isinstance(self._hold_right, list) else [self._hold_right]

            # prepare and hold
            try:
                result = await asyncio.gather(*[
                    self.__prepare(left, self._execute, self._execute_left, hold_left),
                    self.__prepare(right, self._execute, self._execute_right, hold_right)
                ])

                fun = self._fun(*result[0], *result[1])
                return *(await self._execute_fun(fun)), None
            except Exception as e:
                return 0, [str(e)], e



import asyncio
from os import PathLike
from typing import Union, List

from .JPTest import JPTest, EXECUTE_TYPE


class JPTestGet(JPTest):
    """
    decorator to use with test functions
    """

    def __init__(self, name: str = None, max_score: Union[float, int] = None, timeout: int = 120,
                 execute: EXECUTE_TYPE = None, get: Union[str, List[str]] = None):
        super().__init__(name, max_score, timeout, execute)
        self._get: List[str] = get if isinstance(get, list) else [get]

    async def execute(self, notebook: Union[str, PathLike]):
        async with self._start(notebook) as nb:
            try:
                if self._execute is not None:
                    await self._execute_recursively(nb, self._execute)

                references = (nb.ref(name) for name in self._get)
                values = await asyncio.gather(*[ref.receive() for ref in references])

                fun = self._fun(*values)
                return *(await self._execute_fun(fun)), None
            except Exception as e:
                return 0, [str(e)], e



from typing import List


class NotebookError(Exception):
    def __init__(self, name: str, value: str, traceback: List[str]):
        self.name = name
        self.value = value
        self.traceback = traceback



from typing import List, Tuple, Union, Dict, Callable, Awaitable

from nbformat import NotebookNode

from .NotebookError import NotebookError


class NotebookCell:
    CELL_DATA = List[Tuple[str, Union[str, Dict]]]
    CELL_STREAM = Union[str, None]
    # CELL_ERROR = Union[Tuple[str, str, List], None]
    CELL_EXECUTION_RESULT = Tuple[CELL_DATA, CELL_STREAM, CELL_STREAM, CELL_DATA]

    def __init__(self, nb: NotebookNode, ce: Callable[["NotebookCell"], Awaitable], idx: int):
        self._nb: NotebookNode = nb
        self._ce: Callable[[NotebookCell], Awaitable] = ce
        self.idx: int = idx
        self.last_execution_result = None

    @property
    def raw_cell(self) -> NotebookNode:
        return self._nb.cells[self.idx]

    @property
    def type(self) -> str:
        """
        cell type

        :return:
        """
        return self.raw_cell['cell_type']

    @property
    def tags(self) -> List[str]:
        """
        list of tags

        :return:
        """
        if 'tags' in self.raw_cell['metadata']:
            return self.raw_cell['metadata']['tags']
        else:
            return []

    @property
    def source(self) -> str:
        return self.raw_cell['source']

    async def execute(self) -> "NotebookCell":
        """
        execute this cell in notebook context

        :return: self
        """
        self.last_execution_result = await self._ce(self)
        return self

    def output(self) -> CELL_EXECUTION_RESULT:
        """
        extract output data from a cell

        :return: tuple with different output types
        """
        execute_result: NotebookCell.CELL_DATA = []
        stream_stdout: NotebookCell.CELL_STREAM = None
        stream_stderr: NotebookCell.CELL_STREAM = None
        display_data: NotebookCell.CELL_DATA = []
        # error: CELL_ERROR = None

        for o in self.raw_cell['outputs']:
            if o['output_type'] == 'execute_result':
                for k, v in o['data'].items():
                    execute_result.append((k, v))
            elif o['output_type'] == 'stream' and o['name'] == 'stdout':
                stream_stdout = o['text']
            elif o['output_type'] == 'stream' and o['name'] == 'stderr':
                stream_stderr = o['text']
            elif o['output_type'] == 'display_data':
                for k, v in o['data'].items():
                    display_data.append((k, v))
            elif o['output_type'] == 'error':
                raise NotebookError(o['ename'], o['evalue'], o['traceback'])
            else:
                raise AssertionError

        return execute_result, stream_stdout, stream_stderr, display_data  # , error



from .JPSetup import JPSetup, JPTeardown
from .JPTest import JPTest
from .JPTestComparison import JPTestComparison
from .JPTestGet import JPTestGet
from .notebook import Notebook, NotebookReference
from .notebook.kernels import *



import argparse
import asyncio
import importlib.util
import json
import os
import sys
import traceback
from argparse import ArgumentParser
from asyncio import Semaphore
from functools import reduce
from typing import List, Tuple

from jptest2 import JPTest, JPSetup, JPTeardown


async def test(args: argparse.Namespace):
    # override default timeout
    JPTest.DEFAULT_TIMEOUT = args.timeout

    # reset registered tests and other functions
    JPTest.TESTS = {}
    JPSetup.FN = []
    JPTeardown.FN = []

    # load classes from file
    if args.test_file is not None:
        spec = importlib.util.spec_from_file_location('testfile', args.test_file)
        foo = importlib.util.module_from_spec(spec)
        sys.modules[spec.name] = foo
        spec.loader.exec_module(foo)
    else:
        # noinspection PyUnresolvedReferences
        import jptest2.RunTest

    # pre run functions
    if args.verbose:
        print('pre run', file=sys.stderr)

    if len(JPSetup.FN) > 0:
        await asyncio.gather(*[f() for f in JPSetup.FN])

    # filter and execute tests
    proc: Semaphore = Semaphore(args.tests)

    async def exec_test(test: JPTest):
        async with proc:
            return test.max_score, *await test.execute(args.nb_file)

    async def exec_list(name: str, tests: List[JPTest]):
        rs = await asyncio.gather(*[exec_test(test) for test in tests])
        return name, *reduce(
            lambda acc, row: (
                acc[0] + row[0],
                acc[1] + row[1],
                acc[2] + list(set(row[2]) - set(acc[2])),
                acc[3] + [row[3]]
            ),
            rs,
            (0, 0, [], [])
        )

    results: List[Tuple[str, float, float, List[str], List[Exception]]] = await asyncio.gather(*[
        exec_list(name, tests)
        for name, tests in JPTest.TESTS.items()
        if args.test_name is None or args.test_name == name
    ])

    # post run functions
    if args.verbose:
        print('post run', file=sys.stderr)

    if len(JPTeardown.FN) > 0:
        await asyncio.gather(*[f() for f in JPTeardown.FN])

    # print output
    if args.quiet:  # quiet
        for name, max_score, score, _, errors in results:
            for e in errors:
                if e is not None:
                    raise e

            if score != max_score:
                raise AssertionError(f'name={name}, score={score}, max_score={max_score}')

    elif args.md:  # md
        skip_total = (len(results) == 1)

        if not skip_total:
            achieved = sum((score for _, _, score, _, _ in results))
            total = sum((max_score for _, max_score, _, _, _ in results))

            print(f'# Bewertung ({achieved} / {total})')
            print()

            headline = '##'
        else:
            headline = '#'

        for name, max_score, score, comments, _ in results:
            print(f'{headline} {name} ({score} / {max_score})')
            for comment in comments:
                print(f'- {comment}')
            print()

    else:  # json
        achieved = sum((score for _, _, score, _, _ in results))
        total = sum((max_score for _, max_score, _, _, _ in results))

        result = {
            'achievedScore': achieved,
            'totalScore': total,
            'tests': []
        }

        for name, max_score, score, comments, _ in results:
            result['tests'].append({
                'name': name,
                'achievedScore': score,
                'totalScore': max_score,
                'comments': comments
            })

        print(json.dumps(result, indent=4))


async def main():
    # configure argument parser
    parser = ArgumentParser()
    parser.add_argument('nb_file', help='notebook file (.ipynb) to load')
    parser.add_argument('test_file', help='test file (.py) to load', nargs='?')
    parser.add_argument('test_name', help='test to execute (all if None given)', nargs='?')
    parser.add_argument('--json', action='store_true', help='print output as json (default)', default=True)
    parser.add_argument('--tests', type=int, help='number of tests to process concurrently', default=1000)
    parser.add_argument('--md', action='store_true', help='print output as markdown')
    parser.add_argument('--timeout', type=int, help='override default timeout in seconds (default 120s)', default=120)
    parser.add_argument('--quiet', action='store_true', help='only print exceptions')
    parser.add_argument('--verbose', '-v', action='store_true', help='print verbosely to stderr')
    parser.add_argument('--live', action='store_true', help='run infinitely and watch for changes')

    args = parser.parse_args()

    # default mode
    if not args.live:
        await test(args)

    # live mode
    else:
        from watchfiles import awatch

        async def live_test():
            os.system('clear')

            try:
                await test(args)
            except Exception as e:
                traceback.print_exception(e)

        await live_test()

        async for _ in awatch(args.nb_file, args.test_file):
            await live_test()


if __name__ == '__main__':
    asyncio.run(main())



import pytest

from jptest2 import PythonNotebook


@pytest.mark.asyncio
async def test_track_function():
    async with PythonNotebook('functions.ipynb') as nb:
        # execute function definition
        await nb.execute_cells('definition')

        # track function `nb_fun`
        async with nb.track_fun('nb_fun', all_parameters=True, return_values=True) as track:
            await nb.execute_cells('more-calls')

        calls = await track.receive()
        assert len(calls) == 8

        assert calls[0].parameters == [('a', 148), ('b', 116)]
        assert calls[1].parameters == [('a', '179'), ('b', '155')]
        assert calls[2].parameters == [('a', 175.25), ('b', 143.5)]
        assert calls[3].parameters == [('a', True), ('b', False)]
        assert calls[4].parameters == [('a', [5, 3, 1]), ('b', [2, 4, 6])]
        assert calls[5].parameters == [('a', None), ('b', 2048)]
        assert calls[6].parameters == [('a', 5), ('b', 2)]
        assert calls[7].parameters == [('a', 1), ('b', 2)]

        assert calls[0].return_value == 148
        assert calls[1].return_value == '179'
        assert calls[2].return_value == 175.25
        assert calls[3].return_value is True
        assert calls[4].return_value == [5, 3, 1]
        assert calls[5].return_value is None
        assert calls[6].return_value == 5
        assert calls[7].return_value == 1

        # track function `nb_unlimited`
        async with nb.track_fun('nb_unlimited', all_parameters=True, return_values=True) as track:
            await nb.execute_cells('more-calls')

        calls = await track.receive()
        assert len(calls) == 2

        assert calls[0].parameters == [('args', [5, 2])]
        assert calls[1].parameters == [('kwargs', {'a': 1, 'b': 2})]

        assert calls[0].return_value == 5
        assert calls[1].return_value == 1

        # track function `nb1`
        async with nb.track_fun('nb1', all_parameters=True, return_values=True) as track:
            await nb.execute_cells('more-calls')

        calls = await track.receive()
        assert len(calls) == 1

        assert calls[0].parameters == [('b', 1), ('a', 2)]
        assert calls[0].return_value == 2

        # track function `nb2`
        async with nb.track_fun('nb2', all_parameters=True, return_values=True) as track:
            await nb.execute_cells('more-calls')

        calls = await track.receive()
        assert len(calls) == 1

        assert calls[0].parameters == [('a', 3), ('b', 2)]
        assert calls[0].return_value == 2

        # track function `nb3`
        async with nb.track_fun('nb3', all_parameters=True, return_values=True) as track:
            await nb.execute_cells('more-calls')

        calls = await track.receive()
        assert len(calls) == 1

        assert calls[0].parameters == [('a', 1), ('args', [2])]
        assert calls[0].return_value is None

        # track function `nb4`
        async with nb.track_fun('nb4', all_parameters=True, return_values=True) as track:
            await nb.execute_cells('more-calls')

        calls = await track.receive()
        assert len(calls) == 1

        assert calls[0].parameters == [('a', 1), ('b', 2), ('args', [3, 4, 5])]
        assert calls[0].return_value is None

        # track function `nb5`
        async with nb.track_fun('nb5', all_parameters=True, return_values=True) as track:
            await nb.execute_cells('more-calls')

        calls = await track.receive()
        assert len(calls) == 1

        assert calls[0].parameters == [('a', 1), ('kwargs', {'b': 2, 'c': 3})]
        assert calls[0].return_value == {'b': 2, 'c': 3}

        # track function `nb6`
        async with nb.track_fun('nb6', all_parameters=True, return_values=True) as track:
            await nb.execute_cells('more-calls')

        calls = await track.receive()
        assert len(calls) == 1

        assert calls[0].parameters == [('a', 1), ('kwargs', {'c': 3}), ('b', 3)]
        assert calls[0].return_value == 3

        # track function `nb7`
        async with nb.track_fun('nb7', all_parameters=True, return_values=True) as track:
            await nb.execute_cells('more-calls')

        calls = await track.receive()
        assert len(calls) == 1

        assert calls[0].parameters == [('a', 1), ('b', 2)]
        assert calls[0].return_value == 1


@pytest.mark.asyncio
async def test_multiple():
    async with PythonNotebook('functions.ipynb') as nb:
        # execute function definition
        await nb.execute_cells('definition')

        # track functions `nb1` and `nb2`
        async with \
                nb.track_fun('nb1', all_parameters=True, return_values=True) as t1, \
                nb.track_fun('nb2', all_parameters=True, return_values=True) as t2:
            await nb.execute_cells('more-calls')

        # compare output
        calls1 = await t1.receive()
        calls2 = await t2.receive()

        assert len(calls1) == 1
        assert calls1[0].parameters == [('b', 1), ('a', 2)]
        assert calls1[0].return_value == 2

        assert len(calls2) == 1
        assert calls2[0].parameters == [('a', 3), ('b', 2)]
        assert calls2[0].return_value == 2


@pytest.mark.asyncio
async def test_filter():
    async with PythonNotebook('functions.ipynb') as nb:
        # execute function definition
        await nb.execute_cells('definition')

        # track function `nb_fun` with parameter `b`
        async with nb.track_fun('nb_fun', 'b') as track:
            await nb.execute_cells('more-calls')

        calls = await track.receive()
        assert len(calls) == 8

        assert calls[0].parameters == [('b', 116)]
        assert calls[1].parameters == [('b', '155')]
        assert calls[2].parameters == [('b', 143.5)]
        assert calls[3].parameters == [('b', False)]
        assert calls[4].parameters == [('b', [2, 4, 6])]
        assert calls[5].parameters == [('b', 2048)]
        assert calls[6].parameters == [('b', 2)]
        assert calls[7].parameters == [('b', 2)]

        # track function `nb_fun` without any parameters
        async with nb.track_fun('nb_fun') as track:
            await nb.execute_cells('more-calls')

        calls = await track.receive()
        assert len(calls) == 8

        assert calls[0].parameters == []
        assert calls[1].parameters == []
        assert calls[2].parameters == []
        assert calls[3].parameters == []
        assert calls[4].parameters == []
        assert calls[5].parameters == []
        assert calls[6].parameters == []
        assert calls[7].parameters == []


@pytest.mark.asyncio
async def test_return_values_false():
    async with PythonNotebook('functions.ipynb') as nb:
        # execute function definition
        await nb.execute_cells('definition')

        # track function `nb_fun`
        async with nb.track_fun('nb_fun', return_values=False) as track:
            await nb.execute_cells('more-calls')

        calls = await track.receive()
        assert len(calls) == 8

        assert calls[0].return_value is None
        assert calls[1].return_value is None
        assert calls[2].return_value is None
        assert calls[3].return_value is None
        assert calls[4].return_value is None
        assert calls[5].return_value is None
        assert calls[6].return_value is None
        assert calls[7].return_value is None


@pytest.mark.asyncio
async def test_clear():
    async with PythonNotebook('functions.ipynb') as nb:
        # execute function definition
        await nb.execute_cells('definition')

        # track function `nb_fun`
        async with nb.track_fun('nb_fun', return_values=False) as track:
            await nb.execute_cells('more-calls')
            calls = await track.receive()
            assert len(calls) == 8

            await track.clear()
            calls = await track.receive()
            assert len(calls) == 0

            await nb.execute_cells('more-calls')
            calls = await track.receive()
            assert len(calls) == 8


@pytest.mark.asyncio
async def test_first_and_last_call():
    async with PythonNotebook('functions.ipynb') as nb:
        # execute function definition
        await nb.execute_cells('definition')

        # track function
        async with nb.track_fun('nb1', all_parameters=True, return_values=True) as track:
            await nb.ref('nb1')(1, 2).receive()
            await nb.ref('nb1')(5, 4).receive()
            await nb.ref('nb1')(11, 12).receive()

        # get first call
        first_call = await track.receive_first()
        assert first_call.parameters == [('a', 1), ('b', 2)]
        assert first_call.return_value == 1

        # get last call
        last_call = await track.receive_last()
        assert last_call.parameters == [('a', 11), ('b', 12)]
        assert last_call.return_value == 11

        # clear
        await track.clear()

        # get first call
        first_call = await track.receive_first()
        assert first_call is None

        # get last call
        last_call = await track.receive_last()
        assert last_call is None


@pytest.mark.asyncio
async def test_ctx_exit():
    async with PythonNotebook('functions.ipynb') as nb:
        # execute function definition
        await nb.execute_cells('definition')

        # track function
        async with nb.track_fun('nb1', all_parameters=True, return_values=True) as track:
            await nb.ref('nb1')(1, 2).receive()
            await nb.ref('nb1')(3, 4).receive()

        await nb.ref('nb1')(5, 6).receive()

        # get result and compare
        calls = await track.receive()
        assert len(calls) == 2

        assert calls[0].parameters == [('a', 1), ('b', 2)]
        assert calls[0].return_value == 1

        assert calls[1].parameters == [('a', 3), ('b', 4)]
        assert calls[1].return_value == 3



from typing import Callable, Any, List, Awaitable


class JPSetup:
    """
    decorator to use with functions that should be executed prior to all tests
    """
    FN: List[Callable[[], Awaitable]] = []

    def __init__(self, f: Callable[[], Any]):
        self.FN.append(f)


class JPTeardown:
    """
    decorator to use with functions that should be executed after all tests
    """
    FN: List[Callable[[], Any]] = []

    def __init__(self, f: Callable[[], Any]):
        self.FN.append(f)



from . import *


@JPTest('run')
async def run(nb: Notebook):
    # iterate over cells
    for cell in nb.cells:
        # find `jp:input` tag
        replace_input = None

        for tag in cell.tags:
            if tag.startswith('jp:input='):
                replace_input = tag[9:]

        # replace input function if necessary
        if replace_input is not None:
            def replacement(_):
                return replace_input

            async with nb.replace_fun('input', replacement):
                await cell.execute()

        # execute cell normally
        else:
            await cell.execute()



from typing import Tuple, List, Dict, Any, Optional

from . import Notebook


class NotebookFunctionCall:
    """
    stores tracked parameters and results
    """

    def __init__(self, nb: Notebook, fun_name: str,
                 call_parameters: Dict[str, Optional[Any]], return_value: Optional[Any]):
        """
        :param nb:
        :param fun_name: function name
        :param call_parameters: parameter name to parameter value mapping
        :param return_value: functions return value
        """
        self._nb = nb

        self._fun_name = fun_name
        self._parameters: Dict[str, Optional[Any]] = call_parameters
        self._return: Optional[Any] = return_value

    def __getattr__(self, key: str) -> Optional[Any]:
        return self._parameters[key]

    def __getitem__(self, key: str) -> Optional[Any]:
        return self._parameters[key]

    def __contains__(self, key: str) -> bool:
        return key in self._parameters

    @property
    def parameter_names(self) -> List[str]:
        """
        receive all captured parameter names

        :return: list of names
        """
        return list(self._parameters)

    @property
    def parameters(self) -> List[Tuple[str, Optional[Any]]]:
        """
        receive all captured parameter names and values

        :return: list of (name, value)
        """
        return list(self._parameters.items())

    @property
    def return_value(self) -> Optional[Any]:
        """
        receive return value

        :return: return value
        """
        return self._return



import pytest

from jptest2 import PythonNotebook


@pytest.mark.asyncio
async def test_ref():
    async with PythonNotebook('references.ipynb') as nb:
        # create in PythonNotebook
        await nb.execute_cells('create')

        # get references
        a, b, c = nb.refs('a', 'b', 'c')
        nb_swap = nb.ref('nb_swap')

        # compare references
        assert await a.receive() == 'a'
        assert await b.receive() == 'b'
        assert await c.receive() == {
            'a': 1,
            'b': 2
        }
        assert await nb_swap(a, b).receive() == ('b', 'a')

        # change in PythonNotebook
        await nb.execute_cells('change')

        # compare references
        assert await a.receive() == 1
        assert await b.receive() == 'b'
        assert await c.receive() == {
            'a': 11,
            'b': 2
        }
        assert await nb_swap(a, b).receive() == 1


@pytest.mark.asyncio
async def test_get():
    async with PythonNotebook('references.ipynb') as nb:
        # create in PythonNotebook
        await nb.execute_cells('create')

        # get references
        a, b, c = await nb.gets('a', 'b', 'c')
        nb_swap = await nb.get('nb_swap')

        # compare references
        assert await a.receive() == 'a'
        assert await b.receive() == 'b'
        assert await c.receive() == {
            'a': 1,
            'b': 2
        }
        assert await nb_swap(a, b).receive() == ('b', 'a')

        # change in PythonNotebook
        await nb.execute_cells('change')

        # compare references
        assert await a.receive() == 'a'
        assert await b.receive() == 'b'
        assert await c.receive() == {
            'a': 11,
            'b': 2
        }
        assert await nb_swap(a, b).receive() == ('b', 'a')



from uuid import uuid4


# randomize name
def randomize_name(name: str) -> str:
    """
    add a random suffix to a name

    :param name:
    :return: name with random suffix
    """
    name = name.replace('.', '_')
    random_id = str(uuid4()).replace('-', '_')

    return f'_{name}_{random_id}'



from os import PathLike
from typing import Callable, List, Union, Awaitable, Iterator

import nbformat
from nbformat import NotebookNode
from nbformat.v4 import new_code_cell

from .NotebookCell import NotebookCell


class Notebook:
    """
    class for async interaction with notebooks
    """

    def __init__(self, notebook: Union[str, PathLike], ce: Callable[[NotebookCell], Awaitable], execute: bool):
        """
        :param notebook: notebook path
        :param execute: execute all cells in `__aenter__`
        :param execute: timeout per cell in seconds
        """
        self._nb: NotebookNode = nbformat.read(notebook, as_version=4)
        self._ce: Callable[[NotebookCell], Awaitable] = ce
        self._execute: bool = execute

    async def __aenter__(self) -> "Notebook":
        """
        initialize and setup kernel

        :return: self
        """
        if self._execute:
            await self.execute_cells()

        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        """
        clear kernel

        :param exc_type:
        :param exc_val:
        :param exc_tb:
        """
        pass

    async def execute_code(self, code: str) -> NotebookCell:
        """
        execute code in notebook context

        :param code: code string
        :return: cell output after execution
        """
        # create new code cell
        cell = new_code_cell(code)

        # insert into notebook
        insert_index = len(self._nb.cells)
        self._nb.cells.append(cell)

        # execute cell
        nb_cell = NotebookCell(self._nb, self._ce, insert_index)
        await nb_cell.execute()

        # parse output
        return nb_cell

    @property
    def cells(self) -> List[NotebookCell]:
        """
        get a list of all cells

        :return: list of NotebookCell
        """
        return [NotebookCell(self._nb, self._ce, i) for i in range(len(self._nb.cells))]

    def iter_code_cells(self, *tag: str, from_tag: str = None, to_tag: str = None) -> Iterator[NotebookCell]:
        """
        get code cells
          - all cells by default
          - matching at least one of the given tags if given
          - only between first occurrence of `from_tag` to first occurrence of `to_tag` if given

        :param tag: cell tags
        :param from_tag: start execution after first occurrence
        :param to_tag: stop execution after first occurrence
        :return: generator of NotebookCell
        """
        start_tag_found = False
        end_tag_found = False

        for cell in self.cells:
            if from_tag in cell.tags:
                start_tag_found = True
            if to_tag in cell.tags:
                end_tag_found = True

            if cell.type != 'code' or (from_tag is not None and not start_tag_found):
                continue

            if len(tag) == 0 or any([t in cell.tags for t in tag]):
                yield cell

            if end_tag_found:
                break

    def get_code_cells(self, *tag: str, from_tag: str = None, to_tag: str = None) -> List[NotebookCell]:
        """
        get code cells
          - all cells by default
          - matching at least one of the given tags if given
          - only between first occurrence of `from_tag` to first occurrence of `to_tag` if given

        :param tag: cell tags
        :param from_tag: start execution after first occurrence
        :param to_tag: stop execution after first occurrence
        :return: list of NotebookCell
        """
        return list(self.iter_code_cells(*tag, from_tag=from_tag, to_tag=to_tag))

    def get_code_cell(self, *tag: str, from_tag: str = None, to_tag: str = None) -> NotebookCell:
        """
        get the first code cell
          - all cells by default
          - matching at least one of the given tags if given
          - only between first occurrence of `from_tag` to first occurrence of `to_tag` if given

        :param tag: cell tags
        :param from_tag: start execution after first occurrence
        :param to_tag: stop execution after first occurrence
        :return: NotebookCell
        """
        for cell in self.iter_code_cells(*tag, from_tag=from_tag, to_tag=to_tag):
            return cell

        raise ValueError('cell not found')

    async def execute_cells(self, *tag: str, from_tag: str = None, to_tag: str = None) -> List[NotebookCell]:
        """
        execute code cells using `get_code_cells`

        :param tag: cell tags
        :param from_tag: start execution after first occurrence
        :param to_tag: stop execution after first occurrence
        :return: list of NotebookCell
        """
        result: List[NotebookCell] = self.get_code_cells(*tag, from_tag=from_tag, to_tag=to_tag)

        for cell in result:
            await cell.execute()

        return result

    async def execute_cell(self, *tag: str, from_tag: str = None, to_tag: str = None) -> NotebookCell:
        """
        execute the first code cell using `get_code_cells`

        :param tag: cell tags
        :param from_tag: start execution after first occurrence
        :param to_tag: stop execution after first occurrence
        :return: NotebookCell
        """
        cell = self.get_code_cell(*tag, from_tag=from_tag, to_tag=to_tag)

        await cell.execute()
        return cell



import asyncio
import re
from concurrent.futures import ThreadPoolExecutor
from os import PathLike
from typing import Union, List, Tuple

import duckdb

from .. import Notebook
from ..NotebookCell import NotebookCell


class DuckDBNotebook(Notebook):
    def __init__(self, notebook: Union[str, PathLike], execute: bool = False, executor: ThreadPoolExecutor = None):
        super().__init__(notebook, self.__execute_cell, execute)

        self.db: duckdb.DuckDBPyConnection
        self.executor: ThreadPoolExecutor

        if executor is None:
            self.executor = ThreadPoolExecutor(max_workers=1)
        else:
            self.executor = executor

    async def __aenter__(self) -> "Notebook":
        self.db: duckdb.DuckDBPyConnection = duckdb.connect(':memory:')

        self.db.__enter__()
        await super().__aenter__()

        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await super().__aexit__(exc_type, exc_val, exc_tb)
        self.db.__exit__(exc_type, exc_val, exc_tb)

    def _execute_and_fetch(self, statement: str, fetch: bool) -> Tuple[List, List]:
        with self.db.cursor() as cursor:
            cursor.execute(statement)

            if fetch:
                return [c[0] for c in cursor.description], cursor.fetchall()

    async def __execute_cell(self, cell: NotebookCell):
        statements = list(filter(lambda x: x.strip(), re.split(r';$', cell.source, flags=re.MULTILINE)))
        last_index = len(statements) - 1

        for i, statement in enumerate(statements):
            if i == last_index:
                return await asyncio.get_event_loop().run_in_executor(
                    self.executor,
                    self._execute_and_fetch, statement, True
                )
            else:
                await asyncio.get_event_loop().run_in_executor(
                    self.executor,
                    self._execute_and_fetch, statement, False
                )



import base64

import pytest
from nbclient.exceptions import CellExecutionError

from jptest2 import PythonNotebook


def read_image(path: str) -> str:
    with open(path, 'rb') as file:
        data = file.read()
        data = base64.b64encode(data).decode('ascii')

        return data


@pytest.mark.asyncio
async def test_execute_code():
    async with PythonNotebook('execute_code.ipynb') as nb:
        # execute result
        exec, out, err, dsp = (await nb.execute_code('''
            1 + 1
            5 + 15
        ''')).output()

        assert exec == [('text/plain', '20')]
        assert out is None
        assert err is None
        assert dsp == []

        # stdout
        exec, out, err, dsp = (await nb.execute_code('''
            print(2)
            print(5)
        ''')).output()

        assert exec == []
        assert out == '2\n5\n'
        assert err is None
        assert dsp == []

        # stderr
        exec, out, err, dsp = (await nb.execute_code('''
            import sys
            print(3, file=sys.stderr)
            print(6, file=sys.stderr)
        ''')).output()

        assert exec == []
        assert out is None
        assert err == '3\n6\n'
        assert dsp == []

        # exception
        with pytest.raises(CellExecutionError):
            await nb.execute_code('''
                assert 1 == 2
            ''')

        # display data
        exec, out, err, dsp = (await nb.execute_code('''
            from IPython.display import display, Image
            display(Image('stripes.png'))
        ''')).output()

        assert exec == []
        assert out is None
        assert err is None
        assert dsp == [
            ('image/png', f'{read_image("stripes.png")}\n'),
            ('text/plain', '<IPython.core.display.Image object>')
        ]

        # display_data in execute_result
        exec, out, err, dsp = (await nb.execute_code('''
            from IPython.display import display, Image
            Image('stripes.png')
        ''')).output()

        assert exec == [
            ('image/png', f'{read_image("stripes.png")}\n'),
            ('text/plain', '<IPython.core.display.Image object>')
        ]
        assert out is None
        assert err is None
        assert dsp == []


@pytest.mark.asyncio
async def test_store():
    async with \
            PythonNotebook('execute_code.ipynb') as nb, \
            PythonNotebook('execute_code.ipynb') as nb2:
        # store
        val = {
            'a': 1,
            'b': [2, 'c']
        }
        ref = await nb.store(val, 'new_value')
        assert ref.name == 'new_value' and await ref.receive() == val

        # stores
        val1 = [1, 2, 3]
        val2 = {'ex': 'ample'}

        ref1, ref2 = await nb.stores(v1=val1, v2=val2)
        assert await ref1.receive() == val1
        assert await ref2.receive() == val2

        # store reference
        ref3 = await nb.store(ref1)
        assert await ref3.receive() == val1

        ref4 = await nb.store(ref2, 'custom_name')
        assert ref4.name == 'custom_name' and await ref4.receive() == val2

        # store reference from second notebook
        val3 = [5, 4, 3, 2, 1]
        ref5 = await nb2.store(val3)

        ref6 = await nb.store(ref5)
        assert await ref6.receive() == val3


@pytest.mark.asyncio
async def test_execute_all_cells():
    async with PythonNotebook('execute_code.ipynb') as nb:
        cells = await nb.execute_cells()
        assert len(cells) == 10

        result = [c.output() for c in cells]
        assert result == [(
            [],
            None,
            None,
            []
        ), (
            [('text/plain', '6')],
            None,
            None,
            []
        ), (
            [('image/png', f'{read_image("stripes.png")}\n'), ('text/plain', '<IPython.core.display.Image object>')],
            None,
            None,
            []
        ), (
            [('text/plain', '106')],
            None,
            None,
            []
        ), (
            [('text/plain', '111')],
            None,
            None,
            []
        ), (
            [('text/plain', '115')],
            None,
            None,
            []
        ), (
            [('text/plain', '1115')],
            None,
            None,
            []
        ), (
            [('text/plain', '200')],
            None,
            None,
            []
        ), (
            [('text/plain', '201')],
            None,
            None,
            []
        ), (
            [('text/plain', '300')],
            None,
            None,
            []
        )]


@pytest.mark.asyncio
async def test_execute_cells_by_tag():
    async with PythonNotebook('execute_code.ipynb') as nb:
        # only cells with tag 'task-1'
        cells = await nb.execute_cells('task-1')
        assert len(cells) == 2

        result = [c.output() for c in cells]

        assert result == [(
            [('text/plain', '6')],
            None,
            None,
            []
        ), (
            [('text/plain', '10')],
            None,
            None,
            []
        )]

        # only cells with tag `task-2` and `task-3`
        cells = await nb.execute_cells('task-2', 'task-3')
        assert len(cells) == 3

        result = [c.output() for c in cells]

        assert result == [(
            [('text/plain', '106')],
            None,
            None,
            []
        ), (
            [('text/plain', '111')],
            None,
            None,
            []
        ), (
            [('text/plain', '1111')],
            None,
            None,
            []
        )]


@pytest.mark.asyncio
async def test_execute_cells_from_to():
    async with PythonNotebook('execute_code.ipynb') as nb:
        cells = await nb.execute_cells(from_tag='task-2', to_tag='task-4')
        assert len(cells) == 5

        result = [c.output() for c in cells]

        assert result == [(
            [('text/plain', '106')],
            None,
            None,
            []
        ), (
            [('text/plain', '111')],
            None,
            None,
            []
        ), (
            [('text/plain', '115')],
            None,
            None,
            []
        ), (
            [('text/plain', '1115')],
            None,
            None,
            []
        ), (
            [('text/plain', '200')],
            None,
            None,
            []
        )]


@pytest.mark.asyncio
async def test_execute_cells_mixed():
    async with PythonNotebook('execute_code.ipynb') as nb:
        cells = await nb.execute_cells('task-2', 'task-3', from_tag='task-2', to_tag='task-4')
        assert len(cells) == 3

        result = [c.output() for c in cells]

        assert result == [(
            [('text/plain', '106')],
            None,
            None,
            []
        ), (
            [('text/plain', '111')],
            None,
            None,
            []
        ), (
            [('text/plain', '1111')],
            None,
            None,
            []
        )]



import pytest

from jptest2 import PythonNotebook


@pytest.mark.asyncio
async def test_execute_function():
    # simple function
    # noinspection PyUnreachableCode
    def my_fun():
        fun_value = 1
        return fun_value + 1
        return fun_value + 10

    async with PythonNotebook('execute_code.ipynb') as nb:
        result = await nb.execute_fun(my_fun)
        assert result.output() == (
            [('text/plain', '2')],
            None,
            None,
            []
        )

        assert await nb.ref('fun_value').receive() == 1

    # more complex function
    # noinspection PyUnreachableCode
    def my_second_fun():
        fun_value = 1

        if fun_value == 2:
            fun_value = fun_value + 100
            return fun_value
            return fun_value + 10
        else:
            fun_value = fun_value + 1000
            return fun_value
            return fun_value + 10

    async with PythonNotebook('execute_code.ipynb') as nb:
        result = await nb.execute_fun(my_second_fun)
        assert result.output() == (
            [],
            None,
            None,
            []
        )

        assert await nb.ref('fun_value').receive() == 1001


@pytest.mark.asyncio
async def test_execute_inner_function():
    def outer_fun():
        def inner_fun():
            a = 1
            return a + 1

        inner_val = inner_fun()
        return inner_val + 2

    async with PythonNotebook('execute_code.ipynb') as nb:
        result = await nb.execute_fun(outer_fun)
        assert result.output() == (
            [('text/plain', '4')],
            None,
            None,
            []
        )

        assert await nb.ref('inner_val').receive() == 2


@pytest.mark.asyncio
async def test_inject_function():
    def my_fun():
        return 'abc', 1, 'def'

    async with PythonNotebook('execute_code.ipynb') as nb:
        nb_fun = await nb.inject_fun(my_fun)
        result = await nb_fun().receive()

        assert result == my_fun()


@pytest.mark.asyncio
async def test_inject_lambda_function():
    my_lambda = lambda x: x * x

    async with PythonNotebook('execute_code.ipynb') as nb:
        nb_fun = await nb.inject_fun(my_lambda)
        result = await nb_fun(12).receive()

        assert result == my_lambda(12)



import asyncio
from os import PathLike
from types import FunctionType
from typing import List, Tuple, Dict, Callable, AsyncIterable, Awaitable, AsyncGenerator, Iterable
from typing import Union, Optional, Protocol

import aiofiles

from .notebook import Notebook
from .notebook.kernels import *


class JPTestFunction(Protocol):
    def __call__(self, *args, **kwargs) -> Union[Awaitable, AsyncIterable]:
        ...


EXECUTE_TYPE = Union[Tuple[str], Tuple[str, str], str, Callable, PathLike, List['EXECUTE_TYPE']]


class JPTest:
    """
    decorator to use with test functions
    """
    TESTS: Dict[str, List['JPTest']] = {}
    DEFAULT_TIMEOUT = 120

    def __init__(self, name: str = None, max_score: Union[float, int] = 0, timeout: int = None,
                 execute: EXECUTE_TYPE = None, prepare_second: bool = False,
                 kernel: Optional[str] = 'python3'):
        """
        :param name: name used in the output
        :param max_score: maximum score (can be exceeded, used to calculate total score)
        :param timeout: execution timeout in seconds (default: 2 minutes)
        :param execute: cells, code and functions to execute prior to the test
        :param prepare_second: create two notebooks and use `execute` with both in parallel
        """

        self.name: Optional[str] = name
        self.max_score: float = float(max_score)
        self.timeout: int = timeout or JPTest.DEFAULT_TIMEOUT
        self.prepare_second: bool = prepare_second
        self.kernel = kernel

        self._fun: JPTestFunction
        self._execute = execute if execute is not None else []

    def __call__(self, fun: JPTestFunction):
        self._fun: JPTestFunction = fun

        if self.name not in self.TESTS:
            self.TESTS[self.name] = []

        self.TESTS[self.name].append(self)

        return fun

    @property
    def test_name(self) -> Optional[str]:
        return self._fun.__name__

    def _start(self, notebook: Union[str, PathLike]):
        if self.kernel == 'python3':
            return PythonNotebook(notebook, timeout=self.timeout)
        if self.kernel == 'duckdb':
            return DuckDBNotebook(notebook)
        if self.kernel == 'sqlite':
            return SQLiteNotebook(notebook)

        raise AssertionError(f'kernel {self.kernel} not supported')

    @staticmethod
    async def _execute_recursively(nb: Notebook, item: EXECUTE_TYPE):
        """
        :param nb: Notebook object to use
        :param item: item to execute
        :return:
        """

        # tuples
        if isinstance(item, tuple):
            if len(item) == 1:
                await nb.execute_cells(item[0])
            elif len(item) == 2:
                await nb.execute_cells(from_tag=item[0], to_tag=item[1])
            else:
                raise ValueError(f'unsupported tuple length {len(item)}')

        # string
        elif isinstance(item, str):
            await nb.execute_code(item)

        # function
        elif isinstance(item, FunctionType):
            await nb.execute_fun(item)

        # pathlike
        elif isinstance(item, PathLike):
            async with aiofiles.open(item, 'r') as file:
                file_contents = await file.read()

            await nb.execute_code(file_contents)

        # list
        elif isinstance(item, list):
            for i in item:
                await JPTest._execute_recursively(nb, i)

        # other
        else:
            raise ValueError(f'unsupported parameter type {type(item)}')

    async def _execute_fun(self, fun):
        test_score = 0.0
        test_comments = []

        if isinstance(fun, AsyncGenerator):
            async for value in fun:
                if not isinstance(value, tuple):
                    val = True
                    score = value
                    pos_comment = None
                    neg_comment = None
                elif len(value) == 1:
                    val = True
                    score = value[0]
                    pos_comment = None
                    neg_comment = None
                elif len(value) == 2:
                    val, score = value
                    pos_comment = None
                    neg_comment = None
                elif len(value) == 3:
                    val, score, neg_comment = value
                    pos_comment = None
                elif len(value) == 4:
                    val, score, neg_comment, pos_comment = value
                else:
                    raise ValueError(f'invalid yield from test {self.name}')

                if (not isinstance(val, Iterable) and val) or (isinstance(val, Iterable) and all(val)):
                    test_score += score
                    if pos_comment is not None:
                        test_comments.append(pos_comment)
                else:
                    if neg_comment is not None:
                        test_comments.append(neg_comment)

        else:
            await fun

        return test_score, test_comments

    async def execute(self, notebook: Union[str, PathLike]):
        try:
            if not self.prepare_second:
                async with self._start(notebook) as nb:
                    if self._execute is not None:
                        await self._execute_recursively(nb, self._execute)

                    fun = self._fun(nb)
                    return *(await self._execute_fun(fun)), None

            else:
                async with \
                        self._start(notebook) as left, \
                        self._start(notebook) as right:
                    if self._execute is not None:
                        await asyncio.gather(*[
                            self._execute_recursively(left, self._execute),
                            self._execute_recursively(right, self._execute)
                        ])

                    fun = self._fun(left, right)
                    return *(await self._execute_fun(fun)), None

        except Exception as e:
            return 0, [str(e)], e



from typing import List, Set, Optional

from . import Notebook
from .NotebookFunctionCall import NotebookFunctionCall
from .NotebookReference import NotebookReference
from .util import randomize_name


class NotebookFunctionWrapper:
    """
    wraps a function and allows to receive tracked parameters
    """

    def __init__(self, nb: Notebook, fun_name: str,
                 *parameters: str, all_parameters: bool = False, return_values: bool = True):
        """
        :param nb:
        :param fun_name: function name to replace
        :param parameters: parameter names to track
        :param all_parameters: include all parameters (ignore parameters)
        :param return_values: include return values (None otherwise)
        """
        self._nb = nb
        self._fun_name = fun_name

        self._parameters: Set[str] = set(parameters)
        self._all_parameters: bool = all_parameters
        self._return_values: bool = return_values

    async def __aenter__(self) -> "NotebookFunctionWrapper":
        """
        replace function with wrapper

        :return: self
        """
        # store backup
        self._backup: NotebookReference = await self._nb.get(self._fun_name)

        # inject tracker and store reference
        class_name = randomize_name('Track')
        wrapper_name = randomize_name('track')

        await self._nb.execute_code(f'''
            class {class_name}:
                def __init__(self, fun):
                    # store properties
                    self._include_parameters = {self._parameters}
                    self._include_all_parameters = {self._all_parameters}
                    self._include_return_values = {self._return_values}
            
                    # analyse `fun`
                    self.fun = fun
            
                    self._fun_default_values = {{}}
                    self._fun_positions = []
                    self._fun_names = set()
                    self._fun_var_positional = None
                    self._fun_var_keyword = None
            
                    import inspect
                    from inspect import Signature, Parameter
                    signature: Signature = inspect.signature(fun)
            
                    for index, param in enumerate(signature.parameters.values()):
                        if param.default is not Parameter.empty:
                            self._fun_default_values[param.name] = param.default
                        if param.kind == Parameter.POSITIONAL_OR_KEYWORD or param.kind == Parameter.POSITIONAL_ONLY:
                            self._fun_positions.append(param.name)
                        if param.kind == Parameter.POSITIONAL_OR_KEYWORD or param.kind == Parameter.KEYWORD_ONLY:
                            self._fun_names.add(param.name)
                        if param.kind == Parameter.VAR_POSITIONAL:
                            self._fun_var_positional = index
                            self._fun_positions.append(param.name)
                        if param.kind == Parameter.VAR_KEYWORD:
                            self._fun_var_keyword = param.name
            
                    # store calls array
                    self.calls = []
            
                def __call__(self):
                    def wrapper_fun(*args, **kwargs):
                        call_params = {{}}
            
                        # extract args
                        for i, val in enumerate(args):
                            name = self._fun_positions[min(i, len(self._fun_positions) - 1)]
        
                            if self._fun_var_positional is not None and i >= self._fun_var_positional:
                                if name not in call_params:
                                    call_params[name] = []
        
                                call_params[name].append(val)
        
                            else:
                                call_params[name] = val
            
                        # extract kwargs
                        for name, val in kwargs.items():
                            if name in self._fun_names:
                                call_params[name] = val
                            elif self._fun_var_keyword is not None:
                                if self._fun_var_keyword not in call_params:
                                    call_params[self._fun_var_keyword] = {{}}
            
                                call_params[self._fun_var_keyword][name] = val
            
                        # insert default values
                        for name, val in self._fun_default_values.items():
                            if name not in call_params:
                                call_params[name] = val
            
                        # apply filter
                        if not self._include_all_parameters:
                            for key in list(call_params):
                                print(key, self._include_parameters)
                                if key not in self._include_parameters:
                                    del call_params[key]
            
                        # get result
                        result = self.fun(*args, **kwargs)
            
                        # append output
                        if self._include_return_values:
                            self.calls.append((call_params, result))
                        else:
                            self.calls.append((call_params, None))
            
                        # return fun result
                        return result
            
                    return wrapper_fun
            
                def clear(self):
                    self.calls = []
            
            {wrapper_name} = {class_name}({self._fun_name})
            {self._fun_name} = {wrapper_name}()
        ''')

        self._wrapper: NotebookReference = self._nb.ref(wrapper_name)

        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        """
        replace wrapper with original function

        :param exc_type:
        :param exc_val:
        :param exc_tb:
        :return:
        """
        await self._nb.execute_code(f'''
            {self._fun_name} = {self._backup.name}
        ''')

    async def clear(self):
        """
        clear stored calls

        :return:
        """
        await self._wrapper.clear().execute()

    async def receive(self) -> List[NotebookFunctionCall]:
        """
        receive a list of all calls

        :return: list of NotebookFunctionCall
        """
        return list(map(
            lambda c: NotebookFunctionCall(self._nb, self._fun_name, *c),
            await self._wrapper.calls.receive()
        ))

    async def receive_first(self) -> Optional[NotebookFunctionCall]:
        """
        receive the first call made (if any)

        :return: NotebookFunctionCall
        """
        if await self._wrapper.calls.len() > 0:
            return NotebookFunctionCall(self._nb, self._fun_name, *(await self._wrapper.calls[0].receive()))

    async def receive_last(self) -> Optional[NotebookFunctionCall]:
        """
        receive the last call made (if any)

        :return: NotebookFunctionCall
        """
        if await self._wrapper.calls.len() > 0:
            return NotebookFunctionCall(self._nb, self._fun_name, *(await self._wrapper.calls[-1].receive()))



import re
from os import PathLike
from typing import Union, Optional

import aiosqlite

from .. import Notebook
from ..NotebookCell import NotebookCell


class SQLiteNotebook(Notebook):
    def __init__(self, notebook: Union[str, PathLike], execute: bool = False):
        super().__init__(notebook, self.__execute_cell, execute)
        self.db: Optional

    async def __aenter__(self) -> "Notebook":
        self.db = aiosqlite.connect(':memory:')

        await self.db.__aenter__()
        await super().__aenter__()

        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await super().__aexit__(exc_type, exc_val, exc_tb)
        await self.db.__aexit__(exc_type, exc_val, exc_tb)

    async def __execute_cell(self, cell: NotebookCell):
        statements = list(filter(lambda x: x, re.split(r';$', cell.source, flags=re.MULTILINE)))
        last_index = len(statements) - 1

        for i, statement in enumerate(statements):
            async with self.db.execute(statement) as cursor:
                if i == last_index:
                    return await cursor.fetchall()



from setuptools import setup, find_packages

with open('README.md', 'r', encoding='utf-8') as file:
    long_description = file.read()

setup(
    name='jptest2',
    version='2.1.5',
    author='Eric Tröbs',
    author_email='eric.troebs@tu-ilmenau.de',
    description='write graded unit tests for Jupyter Notebooks in a few lines of code',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/erictroebs/jptest',
    project_urls={
        'Bug Tracker': 'https://github.com/erictroebs/jptest/issues',
    },
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    package_dir={'': 'src'},
    packages=find_packages(where='src'),
    python_requires='>=3.6',
    install_requires=[
        'jupyter',
        'aiofiles'
    ],
    extras_require={
        'demo': [
            'watchfiles'
        ],
        'sqlite': [
            'aiosqlite'
        ],
        'duckdb': [
            'duckdb',
            'pandas'
        ]
    }
)



import asyncio
import base64
import pickle
from typing import Any, Union

from . import Notebook, NotebookCell
from .util import randomize_name


class NotebookReference:
    """
    represents references to object in notebook
    """

    def __init__(self, parent, name: str = None):
        self._parent: Union[Notebook, NotebookReference] = parent
        self._name: str = name

    @property
    def _nb(self) -> Notebook:
        if isinstance(self._parent, NotebookReference):
            return self._parent._nb
        else:
            return self._parent

    @property
    def name(self) -> str:
        if self._name is None:
            return self._parent.name
        else:
            return self._name

    async def _resolve(self) -> str:
        return self._name

    def is_from(self, nb: Notebook):
        return nb == self._nb

    async def copy(self, name: str = None) -> "NotebookReference":
        """
        copy object to a new name and return a new reference

        :param name: the new name. random one is used if not provided
        :return: NotebookReference
        """
        if name is None:
            name = randomize_name(self.name)

        await self._nb.execute_code(f'{name} = {await self._resolve()}')
        return NotebookReference(self._nb, name)

    def __getitem__(self, key) -> "NotebookItemReference":
        return NotebookItemReference(self, key)

    def __getattr__(self, key) -> "NotebookAttributeReference":
        return NotebookAttributeReference(self, key)

    def __call__(self, *args, **kwargs) -> "NotebookCallReference":
        """
        call function in notebook context

        :param args:
        :param kwargs:
        :return: NotebookReference to result
        """
        return NotebookCallReference(self, *args, **kwargs)

    async def len(self) -> int:
        """
        receive length of referenced object

        :return: `len(obj)`
        """
        return await self._nb.ref(f'len({await self._resolve()})').receive()

    async def execute(self) -> NotebookCell:
        """
        execute the underlying statement in the notebook context

        :return: inserted notebook cell
        """
        return await self._nb.execute_code(f'''
            import pickle
            {await self._resolve()}
        ''')

    @staticmethod
    async def execute_many(*references: "NotebookReference"):
        return await asyncio.gather(*[ref.execute() for ref in references])

    async def receive(self, deserialize: bool = True) -> Any:
        """
        serialize, transfer and deserialize referenced object from notebook context

        :return: value
        """
        result, o, e, p = (await self._nb.execute_code(f'''
            import pickle, base64
            base64.b64encode(pickle.dumps({await self._resolve()})).decode('ascii')
        ''')).output()

        for mime, value in result:
            if mime != 'text/plain':
                continue

            value = base64.b64decode(value.encode('ascii'))

            if deserialize:
                value = pickle.loads(value)

            return value

    @staticmethod
    async def receive_many(*references: "NotebookReference"):
        return await asyncio.gather(*[ref.receive() for ref in references])


class NotebookItemReference(NotebookReference):
    def __init__(self, parent, key: Any):
        super().__init__(parent)
        self._key: Any = key

    async def _encode(self):
        return pickle.dumps(self._key)

    async def _resolve(self) -> str:
        val, key = await asyncio.gather(
            self._parent._resolve(),
            self._encode()
        )

        return f'{val}[pickle.loads({key})]'


class NotebookAttributeReference(NotebookReference):
    def __init__(self, parent, key: str):
        super().__init__(parent)
        self._key: str = key

    async def _resolve(self) -> str:
        val = await self._parent._resolve()
        return f'{val}.{self._key}'


class NotebookCallReference(NotebookReference):
    def __init__(self, parent, *args, **kwargs):
        super().__init__(parent)
        self._args = args
        self._kwargs = kwargs

    async def _encode_arg(self, arg) -> str:
        # transfer references if not from same notebook
        if isinstance(arg, NotebookReference):
            if not arg.is_from(self._nb):
                arg = await self._nb.store(arg)
        # store local values in notebook
        else:
            arg = await self._nb.store(arg)

        return await arg._resolve()

    async def _encode_kwarg(self, key: str, arg) -> str:
        val = await self._encode_arg(arg)
        return f'{key}={val}'

    async def _resolve(self) -> str:
        # collect parameters
        call_args = []

        for v in self._args:
            call_args.append(self._encode_arg(v))
        for k, v in self._kwargs.items():
            call_args.append(self._encode_kwarg(k, v))

        # create parameter string and return reference
        parent, *call_args = await asyncio.gather(self._parent._resolve(), *call_args)

        call_str = ',\n                '.join(call_args)
        return f'{parent}({call_str})'



from .Notebook import Notebook
from .NotebookCell import NotebookCell
from .NotebookFunctionCall import NotebookFunctionCall
from .NotebookFunctionReplacement import NotebookFunctionReplacement
from .NotebookFunctionWrapper import NotebookFunctionWrapper
from .NotebookReference import NotebookReference



import asyncio
import pickle
from asyncio import Lock
from inspect import getsource
from os import PathLike
from typing import Callable, Tuple, Union, Optional, Any

from nbclient import NotebookClient

from .. import Notebook
from ..NotebookCell import NotebookCell
from ..NotebookFunctionReplacement import NotebookFunctionReplacement
from ..NotebookFunctionWrapper import NotebookFunctionWrapper
from ..NotebookReference import NotebookReference
from ..util import randomize_name


class PythonNotebook(Notebook):
    def __init__(self, notebook: Union[str, PathLike], execute: bool = False, timeout: int = 120):
        super().__init__(notebook, self.__execute_cell, execute)

        self._nc: NotebookClient = NotebookClient(self._nb, kernel_name='python3', timeout=timeout)
        self._lock: Lock = Lock()

    async def __aenter__(self) -> "Notebook":
        """
        initialize kernel

        :return: self
        """
        await self._nc.async_setup_kernel(cleanup_kc=False).__aenter__()
        await super().__aenter__()

        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        """
        clear kernel

        :param exc_type:
        :param exc_val:
        :param exc_tb:
        """
        await super().__aexit__(exc_type, exc_val, exc_tb)
        await self._nc._async_cleanup_kernel()

    async def __execute_cell(self, cell: NotebookCell):
        async with self._lock:
            await self._nc.async_execute_cell(cell.raw_cell, cell_index=cell.idx)

    def __getattr__(self, name: str) -> NotebookReference:
        """
        alias for `ref`

        :param name: name of object
        :return: NotebookReference
        """
        return self.ref(name)

    def ref(self, name: str) -> NotebookReference:
        """
        get a reference to an object inside the notebook context

        :param name: name of object
        :return: NotebookReference
        """
        return NotebookReference(self, name)

    def refs(self, *names: str) -> Tuple[NotebookReference, ...]:
        """
        get multiple references to objects inside the notebook context

        :param names: list of names
        :return: list of NotebookReference
        """
        return tuple(self.ref(n) for n in names)

    async def get(self, name: str) -> NotebookReference:
        """
        copy object to a random name and return a reference

        :param name: name of object
        :return: NotebookReference
        """
        return await self.ref(name).copy()

    async def gets(self, *names: str) -> Tuple[NotebookReference, ...]:
        """
        copy multiple objects to random names and return references

        :param names: list of names
        :return: list of NotebookReference
        """
        return await asyncio.gather(*[self.get(n) for n in names])

    async def store(self, value: Any, name: str = None) -> NotebookReference:
        """
        store a value in the notebook context.
        A random name is chosen if no custom name is given.

        :param value: variable value in notebook
        :param name: variable name in notebook
        :return: reference to created variable
        """
        # choose a random name if no custom name is provided
        if name is None:
            name = randomize_name('stored')

        # handle references
        if isinstance(value, NotebookReference):
            # handle local references (from same notebook)
            if value.is_from(self):
                return await value.copy(name)

            # handle remote references (from other notebook)
            else:
                encoded_value = await value.receive(deserialize=False)

        # handle any other values
        else:
            encoded_value = pickle.dumps(value)

        # send to notebook
        await self.execute_code(f'''
            import pickle
            {name} = pickle.loads({encoded_value})
        ''')

        # return reference
        return self.ref(name)

    async def stores(self, **kwargs) -> Tuple[NotebookReference, ...]:
        """
        store multiple values in the notebook context

        :param kwargs:
        :return: tuple of references
        """
        return await asyncio.gather(*[self.store(v, k) for k, v in kwargs.items()])

    async def execute_fun(self, fun: Callable) -> NotebookCell:
        """
        strip header from function and execute in notebook context

        :param fun: function
        :return:
        """

        # get function code
        code = getsource(fun)

        # split lines
        loc = code.split('\n')

        # strip function header
        loc = loc[1:]

        # remove return statement and subsequent lines
        i: int = 0
        inner_fun: Optional[int] = None
        return_found: Optional[int] = None

        while i < len(loc):
            # strip line and calculate indent
            line = loc[i].lstrip()
            indent = len(loc[i]) - len(line)

            # remove empty lines
            if line == '':
                del loc[i]
                continue

            # compare indent
            if inner_fun is not None and indent <= inner_fun:
                inner_fun = None

            if return_found is not None and indent < return_found:
                return_found = None

            # skip inner functions
            if line.startswith('def ') or line.startswith('async def '):
                inner_fun = indent

            if inner_fun:
                i += 1
                continue

            # remove return statements and lines in block after
            if return_found:
                del loc[i]
                continue

            if line.startswith('return '):
                loc[i] = f'{" " * indent}{line[7:]}'
                return_found = indent

            i += 1

        # recreate code from locs
        code = '\n'.join(loc)

        # execute code
        return await self.execute_code(code)

    async def inject_fun(self, fun: Callable) -> NotebookReference:
        """
        inject function into notebook and return reference

        :param fun: function
        :return:
        """
        code = getsource(fun)
        await self.execute_code(code)

        fun_name = fun.__name__
        if fun_name == '<lambda>':
            fun_name = code.split('=')[0].strip()

        return self.ref(fun_name)

    def replace_fun(self, fun_name: str, replacement: Callable) -> NotebookFunctionReplacement:
        """
        replace a function

        :param fun_name: function to replace in notebook
        :param replacement: function to use as replacement
        :return: instance of NotebookFunctionReplacement
        """
        return NotebookFunctionReplacement(self, fun_name, replacement)

    def track_fun(self, fun_name: str, *parameters: str,
                  all_parameters: bool = False, return_values: bool = True) -> NotebookFunctionWrapper:
        """
        wrap a function to track calls

        :param fun_name: function name
        :param parameters: list of parameters as str (name) or tuple (name, position)
        :param all_parameters: include all parameters (ignore argument `parameters`)
        :param return_values: include return values
        :return: instance of NotebookFunctionWrapper
        """
        return NotebookFunctionWrapper(self, fun_name, *parameters,
                                       all_parameters=all_parameters, return_values=return_values)
